"""Chat Model Components Derived from ChatModel/NVIDIA"""

from __future__ import annotations

import base64
import io
import logging
import os
import sys
import urllib.parse
import warnings
from typing import (
    Any,
    AsyncIterator,
    Callable,
    Dict,
    Iterator,
    List,
    Literal,
    Mapping,
    Optional,
    Sequence,
    Type,
    Union,
)

import requests
from langchain_core._api import deprecated
from langchain_core.callbacks.manager import (
    AsyncCallbackManagerForLLMRun,
    CallbackManagerForLLMRun,
)
from langchain_core.language_models import BaseChatModel, LanguageModelInput
from langchain_core.messages import (
    BaseMessage,
    ChatMessage,
    ChatMessageChunk,
)
from langchain_core.outputs import (
    ChatGeneration,
    ChatGenerationChunk,
    ChatResult,
)
from langchain_core.pydantic_v1 import BaseModel, Field, validator
from langchain_core.runnables import Runnable
from langchain_core.tools import BaseTool

from langchain_nvidia_ai_endpoints import _common as nvidia_ai_endpoints
from langchain_nvidia_ai_endpoints._statics import MODEL_SPECS

_CallbackManager = Union[AsyncCallbackManagerForLLMRun, CallbackManagerForLLMRun]
_DictOrPydanticClass = Union[Dict[str, Any], Type[BaseModel]]
_DictOrPydantic = Union[Dict, BaseModel]

try:
    import PIL.Image

    has_pillow = True
except ImportError:
    has_pillow = False

logger = logging.getLogger(__name__)


def _is_url(s: str) -> bool:
    try:
        result = urllib.parse.urlparse(s)
        return all([result.scheme, result.netloc])
    except Exception as e:
        logger.debug(f"Unable to parse URL: {e}")
        return False


def _resize_image(img_data: bytes, max_dim: int = 1024) -> str:
    if not has_pillow:
        print(  # noqa: T201
            "Pillow is required to resize images down to reasonable scale."
            " Please install it using `pip install pillow`."
            " For now, not resizing; may cause NVIDIA API to fail."
        )
        return base64.b64encode(img_data).decode("utf-8")
    image = PIL.Image.open(io.BytesIO(img_data))
    max_dim_size = max(image.size)
    aspect_ratio = max_dim / max_dim_size
    new_h = int(image.size[1] * aspect_ratio)
    new_w = int(image.size[0] * aspect_ratio)
    resized_image = image.resize((new_w, new_h), PIL.Image.Resampling.LANCZOS)
    output_buffer = io.BytesIO()
    resized_image.save(output_buffer, format="JPEG")
    output_buffer.seek(0)
    resized_b64_string = base64.b64encode(output_buffer.read()).decode("utf-8")
    return resized_b64_string


def _url_to_b64_string(image_source: str) -> str:
    b64_template = "data:image/png;base64,{b64_string}"
    try:
        if _is_url(image_source):
            response = requests.get(image_source)
            response.raise_for_status()
            encoded = base64.b64encode(response.content).decode("utf-8")
            if sys.getsizeof(encoded) > 200000:
                ## (VK) Temporary fix. NVIDIA API has a limit of 250KB for the input.
                encoded = _resize_image(response.content)
            return b64_template.format(b64_string=encoded)
        elif image_source.startswith("data:image"):
            return image_source
        elif os.path.exists(image_source):
            with open(image_source, "rb") as f:
                encoded = base64.b64encode(f.read()).decode("utf-8")
                return b64_template.format(b64_string=encoded)
        else:
            raise ValueError(
                "The provided string is not a valid URL, base64, or file path."
            )
    except Exception as e:
        raise ValueError(f"Unable to process the provided image source: {e}")


class ChatNVIDIA(nvidia_ai_endpoints._NVIDIAClient, BaseChatModel):
    """NVIDIA chat model.

    Example:
        .. code-block:: python

            from langchain_nvidia_ai_endpoints import ChatNVIDIA


            model = ChatNVIDIA(model="meta/llama2-70b")
            response = model.invoke("Hello")
    """

    _default_model: str = "mistralai/mixtral-8x7b-instruct-v0.1"
    infer_endpoint: str = Field("{base_url}/chat/completions")
    model: str = Field(_default_model, description="Name of the model to invoke")
    temperature: Optional[float] = Field(description="Sampling temperature in [0, 1]")
    max_tokens: Optional[int] = Field(
        1024, description="Maximum # of tokens to generate"
    )
    top_p: Optional[float] = Field(description="Top-p for distribution sampling")
    seed: Optional[int] = Field(description="The seed for deterministic results")
    bad: Optional[Sequence[str]] = Field(description="Bad words to avoid (cased)")
    stop: Optional[Sequence[str]] = Field(description="Stop words (cased)")
    labels: Optional[Dict[str, float]] = Field(description="Steering parameters")
    streaming: bool = Field(True)

    @validator("model")
    def aifm_deprecated(cls, value: str) -> str:
        """All AI Foundataion Models are deprecate, use API Catalog models instead."""
        for model in [value, f"playground_{value}"]:
            if model in MODEL_SPECS and MODEL_SPECS[model].get("api_type") == "aifm":
                alternative = MODEL_SPECS[model].get(
                    "alternative", ChatNVIDIA._default_model
                )
                warnings.warn(
                    f"{value} is deprecated. Try {alternative} instead.",
                    DeprecationWarning,
                )
        return value

    @validator("bad")
    def aifm_bad_deprecated(
        cls, value: Optional[Sequence[str]]
    ) -> Optional[Sequence[str]]:
        if value:
            warnings.warn(
                "Bad words are deprecated and not supported by API Catalog models.",
                DeprecationWarning,
            )
        return value

    @validator("labels")
    def aifm_labels_deprecated(
        cls, value: Optional[Sequence[str]]
    ) -> Optional[Sequence[str]]:
        if value:
            warnings.warn(
                "Labels are deprecated and not supported by API Catalog models.",
                DeprecationWarning,
            )
        return value

    @property
    def _llm_type(self) -> str:
        """Return type of NVIDIA AI Foundation Model Interface."""
        return "chat-nvidia-ai-playground"

    def _generate(
        self,
        messages: List[BaseMessage],
        stop: Optional[List[str]] = None,
        run_manager: Optional[CallbackManagerForLLMRun] = None,
        **kwargs: Any,
    ) -> ChatResult:
        inputs = self.custom_preprocess(messages)
        responses = self.get_generation(inputs=inputs, stop=stop, **kwargs)
        self._set_callback_out(responses, run_manager)
        message = ChatMessage(**self.custom_postprocess(responses))
        generation = ChatGeneration(message=message)
        return ChatResult(generations=[generation], llm_output=responses)

    def _get_filled_chunk(self, **kwargs: Any) -> ChatGenerationChunk:
        """Fill the generation chunk."""
        return ChatGenerationChunk(message=ChatMessageChunk(**kwargs))

    def _stream(
        self,
        messages: List[BaseMessage],
        stop: Optional[Sequence[str]] = None,
        run_manager: Optional[CallbackManagerForLLMRun] = None,
        **kwargs: Any,
    ) -> Iterator[ChatGenerationChunk]:
        """Allows streaming to model!"""
        inputs = self.custom_preprocess(messages)
        for response in self.get_stream(inputs=inputs, stop=stop, **kwargs):
            self._set_callback_out(response, run_manager)
            chunk = self._get_filled_chunk(**self.custom_postprocess(response))
            if run_manager:
                run_manager.on_llm_new_token(chunk.text, chunk=chunk)
            yield chunk

    # todo: remove when get_astream is removed
    @deprecated(since="0.0.15", removal="0.1.0")
    async def _astream(
        self,
        messages: List[BaseMessage],
        stop: Optional[Sequence[str]] = None,
        run_manager: Optional[AsyncCallbackManagerForLLMRun] = None,
        **kwargs: Any,
    ) -> AsyncIterator[ChatGenerationChunk]:
        inputs = self.custom_preprocess(messages)
        async for response in self.get_astream(inputs=inputs, stop=stop, **kwargs):
            self._set_callback_out(response, run_manager)
            chunk = self._get_filled_chunk(**self.custom_postprocess(response))
            if run_manager:
                await run_manager.on_llm_new_token(chunk.text, chunk=chunk)
            yield chunk

    def _set_callback_out(
        self,
        result: dict,
        run_manager: Optional[_CallbackManager],
    ) -> None:
        result.update({"model_name": self.model})
        if run_manager:
            for cb in run_manager.handlers:
                if hasattr(cb, "llm_output"):
                    cb.llm_output = result

    @deprecated(since="0.0.15", removal="0.1.0")
    def custom_preprocess(
        self, msg_list: Sequence[BaseMessage]
    ) -> List[Dict[str, str]]:
        return [self.preprocess_msg(m) for m in msg_list]

    def _process_content(self, content: Union[str, List[Union[dict, str]]]) -> str:
        if isinstance(content, str):
            return content
        string_array: list = []

        for part in content:
            if isinstance(part, str):
                string_array.append(part)
            elif isinstance(part, Mapping):
                # OpenAI Format
                if "type" in part:
                    if part["type"] == "text":
                        string_array.append(str(part["text"]))
                    elif part["type"] == "image_url":
                        img_url = part["image_url"]
                        if isinstance(img_url, dict):
                            if "url" not in img_url:
                                raise ValueError(
                                    f"Unrecognized message image format: {img_url}"
                                )
                            img_url = img_url["url"]
                        b64_string = _url_to_b64_string(img_url)
                        string_array.append(f'<img src="{b64_string}" />')
                    else:
                        raise ValueError(
                            f"Unrecognized message part type: {part['type']}"
                        )
                else:
                    raise ValueError(f"Unrecognized message part format: {part}")
        return "".join(string_array)

    @deprecated(since="0.0.15", removal="0.1.0")
    def preprocess_msg(self, msg: BaseMessage) -> Dict[str, str]:
        if isinstance(msg, BaseMessage):
            role_convert = {"ai": "assistant", "human": "user"}
            if isinstance(msg, ChatMessage):
                role = msg.role
            else:
                role = msg.type
            role = role_convert.get(role, role)
            content = self._process_content(msg.content)
            return {"role": role, "content": content}
        raise ValueError(f"Invalid message: {repr(msg)} of type {type(msg)}")

    @deprecated(since="0.0.15", removal="0.1.0")
    def custom_postprocess(self, msg: dict) -> dict:
        kw_left = msg.copy()
        out_dict = {
            "role": kw_left.pop("role", "assistant") or "assistant",
            "name": kw_left.pop("name", None),
            "id": kw_left.pop("id", None),
            "content": kw_left.pop("content", "") or "",
            "additional_kwargs": {},
            "response_metadata": {},
        }
        for k in list(kw_left.keys()):
            if "tool" in k:
                out_dict["additional_kwargs"][k] = kw_left.pop(k)
        out_dict["response_metadata"] = kw_left
        return out_dict

    ######################################################################################
    ## Core client-side interfaces

    @deprecated(since="0.0.15", removal="0.1.0")
    def get_generation(
        self,
        inputs: Sequence[Dict],
        **kwargs: Any,
    ) -> dict:
        """Call to client generate method with call scope"""
        stop = kwargs["stop"] = kwargs.get("stop") or self.stop
        payload = self.get_payload(inputs=inputs, stream=False, **kwargs)
        out = self.client.get_req_generation(self.model, stop=stop, payload=payload)
        return out

    @deprecated(since="0.0.15", removal="0.1.0")
    def get_stream(
        self,
        inputs: Sequence[Dict],
        **kwargs: Any,
    ) -> Iterator:
        """Call to client stream method with call scope"""
        stop = kwargs["stop"] = kwargs.get("stop") or self.stop
        payload = self.get_payload(inputs=inputs, stream=True, **kwargs)
        return self.client.get_req_stream(self.model, stop=stop, payload=payload)

    @deprecated(since="0.0.15", removal="0.1.0")
    def get_astream(
        self,
        inputs: Sequence[Dict],
        **kwargs: Any,
    ) -> AsyncIterator:
        """Call to client astream methods with call scope"""
        stop = kwargs["stop"] = kwargs.get("stop") or self.stop
        payload = self.get_payload(inputs=inputs, stream=True, **kwargs)
        return self.client.get_req_astream(self.model, stop=stop, payload=payload)

    @deprecated(since="0.0.15", removal="0.1.0")
    def get_payload(self, inputs: Sequence[Dict], **kwargs: Any) -> dict:
        """Generates payload for the _NVIDIAClient API to send to service."""
        attr_kwargs = {
            "temperature": self.temperature,
            "max_tokens": self.max_tokens,
            "top_p": self.top_p,
            "seed": self.seed,
            "bad": self.bad,
            "stop": self.stop,
            "labels": self.labels,
        }
        if model_name := self.get_binding_model():
            attr_kwargs["model"] = model_name
        attr_kwargs = {k: v for k, v in attr_kwargs.items() if v is not None}
        new_kwargs = {**attr_kwargs, **kwargs}
        return self.prep_payload(inputs=inputs, **new_kwargs)

    @deprecated(since="0.0.15", removal="0.1.0")
    def prep_payload(self, inputs: Sequence[Dict], **kwargs: Any) -> dict:
        """Prepares a message or list of messages for the payload"""
        messages = [self.prep_msg(m) for m in inputs]
        if kwargs.get("labels"):
            # (WFH) Labels are currently (?) always passed as an assistant
            # suffix message, but this API seems less stable.
            messages += [{"labels": kwargs.pop("labels"), "role": "assistant"}]
        if kwargs.get("stop") is None:
            kwargs.pop("stop")
        return {"messages": messages, **kwargs}

    @deprecated(since="0.0.15", removal="0.1.0")
    def prep_msg(self, msg: Union[str, dict, BaseMessage]) -> dict:
        """Helper Method: Ensures a message is a dictionary with a role and content."""
        if isinstance(msg, str):
            # (WFH) this shouldn't ever be reached but leaving this here bcs
            # it's a Chesterton's fence I'm unwilling to touch
            return dict(role="user", content=msg)
        if isinstance(msg, dict):
            if msg.get("content", None) is None:
                raise ValueError(f"Message {msg} has no content")
            return msg
        raise ValueError(f"Unknown message received: {msg} of type {type(msg)}")

    def bind_tools(
        self,
        tools: Sequence[Union[Dict[str, Any], Type[BaseModel], Callable, BaseTool]],
        *,
        tool_choice: Optional[Union[dict, str, Literal["auto", "none"], bool]] = None,
        **kwargs: Any,
    ) -> Runnable[LanguageModelInput, BaseMessage]:
        raise NotImplementedError(
            "Not implemented, awaiting server-side function-recieving API"
            " Consider following open-source LLM agent spec techniques:"
            " https://huggingface.co/blog/open-source-llms-as-agents"
        )

    def bind_functions(
        self,
        functions: Sequence[Union[Dict[str, Any], Type[BaseModel], Callable]],
        function_call: Optional[str] = None,
        **kwargs: Any,
    ) -> Runnable[LanguageModelInput, BaseMessage]:
        raise NotImplementedError(
            "Not implemented, awaiting server-side function-recieving API"
            " Consider following open-source LLM agent spec techniques:"
            " https://huggingface.co/blog/open-source-llms-as-agents"
        )

    def with_structured_output(
        self,
        schema: _DictOrPydanticClass,
        *,
        method: Literal["function_calling", "json_mode"] = "function_calling",
        return_type: Literal["parsed", "all"] = "parsed",
        **kwargs: Any,
    ) -> Runnable[LanguageModelInput, _DictOrPydantic]:
        raise NotImplementedError(
            "Not implemented, awaiting server-side function-recieving API"
            " Consider following open-source LLM agent spec techniques:"
            " https://huggingface.co/blog/open-source-llms-as-agents"
        )



from __future__ import annotations

import json
import logging
import os
import time
from copy import deepcopy
from typing import (
    Any,
    AsyncIterator,
    Callable,
    Dict,
    Generator,
    Iterator,
    List,
    Literal,
    Optional,
    Sequence,
    Tuple,
    Union,
)
from urllib.parse import urlparse

import aiohttp
import requests
from langchain_core._api import deprecated, warn_deprecated
from langchain_core.pydantic_v1 import (
    BaseModel,
    Field,
    PrivateAttr,
    SecretStr,
    root_validator,
    validator,
)
from requests.models import Response

from langchain_nvidia_ai_endpoints._statics import MODEL_SPECS, Model

logger = logging.getLogger(__name__)

_MODE_TYPE = Literal["nvidia", "nim"]


def default_payload_fn(payload: dict) -> dict:
    return payload


class NVEModel(BaseModel):

    """
    Underlying Client for interacting with the AI Foundation Model Function API.
    Leveraged by the NVIDIABaseModel to provide a simple requests-oriented interface.
    Direct abstraction over NGC-recommended streaming/non-streaming Python solutions.

    NOTE: Models in the playground does not currently support raw text continuation.
    """

    # todo: add a validator for requests.Response (last_response attribute) and
    #       remove arbitrary_types_allowed=True
    class Config:
        arbitrary_types_allowed = True

    ## Core defaults. These probably should not be changed
    _api_key_var = "NVIDIA_API_KEY"
    base_url: str = Field(
        "https://api.nvcf.nvidia.com/v2/nvcf",
        description="Base URL for standard inference",
    )
    get_session_fn: Callable = Field(requests.Session)
    get_asession_fn: Callable = Field(aiohttp.ClientSession)
    endpoints: dict = Field(
        {
            "infer": "{base_url}/pexec/functions/{model_id}",
            "status": "{base_url}/pexec/status/{request_id}",
            "models": "{base_url}/functions",
        }
    )

    api_key: SecretStr = Field(..., description="API Key for service of choice")

    ## Generation arguments
    timeout: float = Field(60, ge=0, description="Timeout for waiting on response (s)")
    interval: float = Field(0.02, ge=0, description="Interval for pulling response")
    last_inputs: dict = Field({}, description="Last inputs sent over to the server")
    last_response: Response = Field(
        None, description="Last response sent from the server"
    )
    payload_fn: Callable = Field(
        default_payload_fn, description="Function to process payload"
    )
    headers_tmpl: dict = Field(
        ...,
        description="Headers template for API calls."
        " Should contain `call` and `stream` keys.",
    )
    _available_functions: Optional[List[dict]] = PrivateAttr(default=None)
    _available_models: Optional[dict] = PrivateAttr(default=None)

    @classmethod
    def is_lc_serializable(cls) -> bool:
        return True

    @property
    def lc_secrets(self) -> Dict[str, str]:
        return {"api_key": self._api_key_var}

    @property
    def headers(self) -> dict:
        """Return headers with API key injected"""
        headers_ = self.headers_tmpl.copy()
        for header in headers_.values():
            if "{api_key}" in header["Authorization"]:
                header["Authorization"] = header["Authorization"].format(
                    api_key=self.api_key.get_secret_value(),
                )
        return headers_

    @validator("base_url")
    def validate_base_url(cls, v: str) -> str:
        if v is not None:
            result = urlparse(v)
            # Ensure scheme and netloc (domain name) are present
            if not (result.scheme and result.netloc):
                raise ValueError(
                    f"Invalid base_url, minimally needs scheme and netloc: {v}"
                )
        return v

    @root_validator(pre=True)
    def validate_model(cls, values: Dict[str, Any]) -> Dict[str, Any]:
        """Validate and update model arguments, including API key and formatting"""
        values["api_key"] = (
            values.get(cls._api_key_var.lower())
            or values.get("api_key")
            or os.getenv(cls._api_key_var)
            or ""
        )
        if "headers_tmpl" not in values:
            call_kvs = {
                "Accept": "application/json",
            }
            stream_kvs = {
                "Accept": "text/event-stream",
                "content-type": "application/json",
            }
            shared_kvs = {
                "Authorization": "Bearer {api_key}",
                "User-Agent": "langchain-nvidia-ai-endpoints",
            }
            values["headers_tmpl"] = {
                "call": {**call_kvs, **shared_kvs},
                "stream": {**stream_kvs, **shared_kvs},
            }
        return values

    @property
    def available_models(self) -> dict:
        """List the available models that can be invoked."""
        if self._available_models is not None:
            return self._available_models
        live_fns = self.available_functions
        if "status" in live_fns[0]:
            live_fns = [v for v in live_fns if v.get("status") == "ACTIVE"]
            self._available_models = {v["name"]: v["id"] for v in live_fns}
        else:
            self._available_models = {v.get("id"): v.get("owned_by") for v in live_fns}
        return self._available_models

    @property
    def available_functions(self) -> list:
        """List the available functions that can be invoked."""
        if self._available_functions and isinstance(self._available_functions, list):
            return self._available_functions
        if not self.endpoints.get("models"):
            raise ValueError("No models endpoint found, so cannot retrieve model list.")
        try:
            invoke_url = self.endpoints.get("models", "").format(base_url=self.base_url)
            query_res = self.query(invoke_url)
        except Exception as e:
            raise ValueError(f"Failed to query model endpoint {invoke_url}.\n{e}")
        output = None
        if isinstance(query_res.get("functions"), list):
            output = query_res.get("functions")
        elif isinstance(query_res.get("data"), list):
            output = query_res.get("data")
        if not isinstance(output, list):
            raise ValueError(
                f"Unexpected response when querying {invoke_url}\n{query_res}"
            )
        # if there's an alias / model name for the function, add it as well
        # this lets users work with ai-gemma-2b and google/gemma-2b
        aliases = []
        for function in output:
            assert (
                "name" in function or "id" in function
            ), f"No name or id in function: {function}"
            name = function["name"] if "name" in function else function["id"]
            if name in MODEL_SPECS and "model_name" in MODEL_SPECS[name]:
                alias = function.copy()
                alias.update(name=MODEL_SPECS[name]["model_name"])
                aliases.append(alias)
        output.extend(aliases)
        self._available_functions = output
        return self._available_functions

    def reset_method_cache(self) -> None:
        """Reset method cache to force re-fetch of available functions"""
        self._available_functions = None
        self._available_models = None

    ####################################################################################
    ## Core utilities for posting and getting from NV Endpoints

    def _post(
        self,
        invoke_url: str,
        payload: Optional[dict] = {},
    ) -> Tuple[Response, Any]:
        """Method for posting to the AI Foundation Model Function API."""
        self.last_inputs = {
            "url": invoke_url,
            "headers": self.headers["call"],
            "json": self.payload_fn(payload),
            "stream": False,
        }
        session = self.get_session_fn()
        self.last_response = response = session.post(**self.last_inputs)
        self._try_raise(response)
        return response, session

    def _get(
        self,
        invoke_url: str,
        payload: Optional[dict] = {},
    ) -> Tuple[Response, Any]:
        """Method for getting from the AI Foundation Model Function API."""
        self.last_inputs = {
            "url": invoke_url,
            "headers": self.headers["call"],
            "stream": False,
        }
        if payload:
            self.last_inputs["json"] = self.payload_fn(payload)
        session = self.get_session_fn()
        self.last_response = response = session.get(**self.last_inputs)
        self._try_raise(response)
        return response, session

    def _wait(self, response: Response, session: Any) -> Response:
        """Wait for a response from API after an initial response is made"""
        start_time = time.time()
        while response.status_code == 202:
            time.sleep(self.interval)
            if (time.time() - start_time) > self.timeout:
                raise TimeoutError(
                    f"Timeout reached without a successful response."
                    f"\nLast response: {str(response)}"
                )
            request_id = response.headers.get("NVCF-REQID", "")
            endpoint_args = {"base_url": self.base_url, "request_id": request_id}
            self.last_response = response = session.get(
                self.endpoints["status"].format(**endpoint_args),
                headers=self.headers["call"],
            )
        self._try_raise(response)
        return response

    def _try_raise(self, response: Response) -> None:
        """Try to raise an error from a response"""
        try:
            response.raise_for_status()
        except requests.HTTPError:
            try:
                rd = response.json()
                if "detail" in rd and "reqId" in rd.get("detail", ""):
                    rd_buf = "- " + str(rd["detail"])
                    rd_buf = rd_buf.replace(": ", ", Error: ").replace(", ", "\n- ")
                    rd["detail"] = rd_buf
            except json.JSONDecodeError:
                rd = response.__dict__
                if "status_code" in rd:
                    if "headers" in rd and "WWW-Authenticate" in rd["headers"]:
                        rd["detail"] = rd.get("headers").get("WWW-Authenticate")
                        rd["detail"] = rd["detail"].replace(", ", "\n")
                else:
                    rd = rd.get("_content", rd)
                    if isinstance(rd, bytes):
                        rd = rd.decode("utf-8")[5:]  ## remove "data:" prefix
                    try:
                        rd = json.loads(rd)
                    except Exception:
                        rd = {"detail": rd}
            status = rd.get("status") or rd.get("status_code") or "###"
            title = (
                rd.get("title")
                or rd.get("error")
                or rd.get("reason")
                or "Unknown Error"
            )
            header = f"[{status}] {title}"
            body = ""
            if "requestId" in rd:
                if "detail" in rd:
                    body += f"{rd['detail']}\n"
                body += "RequestID: " + rd["requestId"]
            else:
                body = rd.get("detail", rd)
            if str(status) == "401":
                body += "\nPlease check or regenerate your API key."
            # todo: raise as an HTTPError
            raise Exception(f"{header}\n{body}") from None

    ####################################################################################
    ## Simple query interface to show the set of model options

    def query(
        self,
        invoke_url: str,
        payload: Optional[dict] = None,
        request: str = "get",
    ) -> dict:
        """Simple method for an end-to-end get query. Returns result dictionary"""
        if request == "get":
            response, session = self._get(invoke_url, payload)
        else:
            response, session = self._post(invoke_url, payload)
        response = self._wait(response, session)
        output = self._process_response(response)[0]
        return output

    def _process_response(self, response: Union[str, Response]) -> List[dict]:
        """General-purpose response processing for single responses and streams"""
        if hasattr(response, "json"):  ## For single response (i.e. non-streaming)
            try:
                return [response.json()]
            except json.JSONDecodeError:
                response = str(response.__dict__)
        if isinstance(response, str):  ## For set of responses (i.e. streaming)
            msg_list = []
            for msg in response.split("\n\n"):
                if "{" not in msg:
                    continue
                msg_list += [json.loads(msg[msg.find("{") :])]
            return msg_list
        raise ValueError(f"Received ill-formed response: {response}")

    def _get_invoke_url(
        self,
        model_name: Optional[str] = None,
        invoke_url: Optional[str] = None,
        endpoint: str = "",
    ) -> str:
        """Helper method to get invoke URL from a model name, URL, or endpoint stub"""
        if not invoke_url:
            endpoint_str = self.endpoints.get(endpoint, "")
            if not endpoint_str:
                raise ValueError(f"Unknown endpoint referenced {endpoint} provided")
            if "{model_id}" in endpoint_str:
                if not model_name:
                    raise ValueError("URL or model name must be specified to invoke")
                if model_name in self.available_models:
                    model_id = self.available_models[model_name]
                elif f"playground_{model_name}" in self.available_models:
                    model_id = self.available_models[f"playground_{model_name}"]
                else:
                    available_models_str = "\n".join(
                        [f"{k} - {v}" for k, v in self.available_models.items()]
                    )
                    raise ValueError(
                        f"Unknown model name {model_name} specified."
                        "\nAvailable models are:\n"
                        f"{available_models_str}"
                    )
            else:
                model_id = ""

            endpoint_args = {"base_url": self.base_url, "model_id": model_id}
            invoke_url = endpoint_str.format(**endpoint_args)

        if not invoke_url:
            raise ValueError("URL or model name must be specified to invoke")

        return invoke_url

    ####################################################################################
    ## Generation interface to allow users to generate new values from endpoints

    def get_req(
        self,
        model_name: Optional[str] = None,
        payload: dict = {},
        invoke_url: Optional[str] = None,
        stop: Optional[Sequence[str]] = None,
        endpoint: str = "",
    ) -> Response:
        """Post to the API."""
        invoke_url = self._get_invoke_url(model_name, invoke_url, endpoint=endpoint)
        if payload.get("stream", False) is True:
            payload = {**payload, "stream": False}
        response, session = self._post(invoke_url, payload)
        return self._wait(response, session)

    def get_req_generation(
        self,
        model_name: Optional[str] = None,
        payload: dict = {},
        invoke_url: Optional[str] = None,
        stop: Optional[Sequence[str]] = None,
        endpoint: str = "infer",
    ) -> dict:
        """Method for an end-to-end post query with NVE post-processing."""
        invoke_url = self._get_invoke_url(model_name, invoke_url, endpoint=endpoint)
        response = self.get_req(model_name, payload, invoke_url)
        output, _ = self.postprocess(response, stop=stop)
        return output

    def postprocess(
        self, response: Union[str, Response], stop: Optional[Sequence[str]] = None
    ) -> Tuple[dict, bool]:
        """Parses a response from the AI Foundation Model Function API.
        Strongly assumes that the API will return a single response.
        """
        msg_list = self._process_response(response)
        msg, is_stopped = self._aggregate_msgs(msg_list)
        msg, is_stopped = self._early_stop_msg(msg, is_stopped, stop=stop)
        return msg, is_stopped

    def _aggregate_msgs(self, msg_list: Sequence[dict]) -> Tuple[dict, bool]:
        """Dig out relevant details of aggregated message"""
        content_buffer: Dict[str, Any] = dict()
        content_holder: Dict[Any, Any] = dict()
        usage_holder: Dict[Any, Any] = dict()  ####
        is_stopped = False
        for msg in msg_list:
            usage_holder = msg.get("usage", {})  ####
            if "choices" in msg:
                ## Tease out ['choices'][0]...['delta'/'message']
                msg = msg.get("choices", [{}])[0]
                is_stopped = msg.get("finish_reason", "") == "stop"
                msg = msg.get("delta", msg.get("message", msg.get("text", "")))
                if not isinstance(msg, dict):
                    msg = {"content": msg}
            elif "data" in msg:
                ## Tease out ['data'][0]...['embedding']
                msg = msg.get("data", [{}])[0]
            content_holder = msg
            for k, v in msg.items():
                if k in ("content",) and k in content_buffer:
                    content_buffer[k] += v
                else:
                    content_buffer[k] = v
            if is_stopped:
                break
        content_holder = {**content_holder, **content_buffer}
        if usage_holder:
            content_holder.update(token_usage=usage_holder)  ####
        return content_holder, is_stopped

    def _early_stop_msg(
        self, msg: dict, is_stopped: bool, stop: Optional[Sequence[str]] = None
    ) -> Tuple[dict, bool]:
        """Try to early-terminate streaming or generation by iterating over stop list"""
        content = msg.get("content", "")
        if content and stop:
            for stop_str in stop:
                if stop_str and stop_str in content:
                    msg["content"] = content[: content.find(stop_str) + 1]
                    is_stopped = True
        return msg, is_stopped

    ####################################################################################
    ## Streaming interface to allow you to iterate through progressive generations

    def get_req_stream(
        self,
        model: Optional[str] = None,
        payload: dict = {},
        invoke_url: Optional[str] = None,
        stop: Optional[Sequence[str]] = None,
        endpoint: str = "infer",
    ) -> Iterator:
        invoke_url = self._get_invoke_url(model, invoke_url, endpoint=endpoint)
        if payload.get("stream", True) is False:
            payload = {**payload, "stream": True}
        self.last_inputs = {
            "url": invoke_url,
            "headers": self.headers["stream"],
            "json": self.payload_fn(payload),
            "stream": True,
        }
        response = self.get_session_fn().post(**self.last_inputs)
        self._try_raise(response)
        call = self.copy()

        def out_gen() -> Generator[dict, Any, Any]:
            ## Good for client, since it allows self.last_inputs
            for line in response.iter_lines():
                if line and line.strip() != b"data: [DONE]":
                    line = line.decode("utf-8")
                    msg, final_line = call.postprocess(line, stop=stop)
                    yield msg
                    if final_line:
                        break
                self._try_raise(response)

        return (r for r in out_gen())

    ####################################################################################
    ## Asynchronous streaming interface to allow multiple generations to happen at once.

    async def get_req_astream(
        self,
        model: Optional[str] = None,
        payload: dict = {},
        invoke_url: Optional[str] = None,
        stop: Optional[Sequence[str]] = None,
        endpoint: str = "infer",
    ) -> AsyncIterator:
        invoke_url = self._get_invoke_url(model, invoke_url, endpoint=endpoint)
        if payload.get("stream", True) is False:
            payload = {**payload, "stream": True}
        self.last_inputs = {
            "url": invoke_url,
            "headers": self.headers["stream"],
            "json": self.payload_fn(payload),
        }
        async with self.get_asession_fn() as session:
            async with session.post(**self.last_inputs) as response:
                self._try_raise(response)
                async for line in response.content.iter_any():
                    if line and line.strip() != b"data: [DONE]":
                        line = line.decode("utf-8")
                        msg, final_line = self.postprocess(line, stop=stop)
                        yield msg
                        if final_line:
                            break


class _NVIDIAClient(BaseModel):
    """
    Higher-Level AI Foundation Model Function API Client with argument defaults.
    Is subclassed by ChatNVIDIA to provide a simple LangChain interface.
    """

    client: NVEModel = Field(NVEModel)

    _default_model: str = ""
    model: str = Field(description="Name of the model to invoke")
    infer_endpoint: str = Field("{base_url}/chat/completions")
    curr_mode: _MODE_TYPE = Field("nvidia")  # todo: remove this in 0.1
    is_hosted: bool = Field(True)

    def __init__(self, **kwargs: Any):
        super().__init__(**kwargs)
        if "base_url" in kwargs:
            self.is_hosted = False
            self.curr_mode = "nim"
            self.client.endpoints["infer"] = self.infer_endpoint
            self.client.endpoints["models"] = "{base_url}/models"

    ####################################################################################

    @root_validator(pre=True)
    def validate_client(cls, values: Any) -> Any:
        """Validate and update client arguments, including API key and formatting"""
        if not values.get("client"):
            values["client"] = NVEModel(**values)
        elif isinstance(values["client"], NVEModel):
            values["client"] = values["client"].__class__(**values["client"].dict())
        if not values.get("model"):
            values["model"] = cls._default_model
            assert values["model"], "No model given, with no default to fall back on."

        # the only model that doesn't support a stream parameter is kosmos_2.
        # to address this, we'll use the payload_fn to remove the stream parameter for
        # kosmos_2. if a user tries to set their own payload_fn, this patch will be
        # overwritten.
        # todo: get kosmos_2 api updated to support stream parameter
        if values["model"] == "kosmos_2":

            def kosmos_patch(payload: dict) -> dict:
                payload.pop("stream", None)
                return payload

            values["client"].payload_fn = kosmos_patch

        return values

    @classmethod
    def is_lc_serializable(cls) -> bool:
        return True

    @property
    def lc_secrets(self) -> Dict[str, str]:
        return {"api_key": self.client._api_key_var}

    @property
    def lc_attributes(self) -> Dict[str, Any]:
        attributes: Dict[str, Any] = {}
        if getattr(self.client, "base_url"):
            attributes["base_url"] = self.client.base_url

        if self.model:
            attributes["model"] = self.model

        if getattr(self.client, "endpoints"):
            attributes["endpoints"] = self.client.endpoints

        return attributes

    @property
    @deprecated(since="0.0.15", removal="0.1.0", alternative="available_models")
    def available_functions(self) -> List[dict]:
        """Map the available functions that can be invoked."""
        return self.__class__.get_available_functions(client=self)

    @property
    def available_models(self) -> List[Model]:
        """Map the available models that can be invoked."""
        if self.curr_mode == "nim" or not self.is_hosted:
            return self.__class__.get_available_models(
                client=self, base_url=self.client.base_url
            )
        else:
            return self.__class__.get_available_models(client=self)

    @classmethod
    @deprecated(since="0.0.15", removal="0.1.0", alternative="get_available_models")
    def get_available_functions(
        cls,
        mode: Optional[_MODE_TYPE] = None,
        client: Optional[_NVIDIAClient] = None,
        **kwargs: Any,
    ) -> List[dict]:
        """Map the available functions that can be invoked. Callable from class"""
        nveclient = (client or cls(**kwargs)).mode(mode, **kwargs).client
        nveclient.reset_method_cache()
        return nveclient.available_functions

    @classmethod
    def get_available_models(
        cls,
        mode: Optional[_MODE_TYPE] = None,
        client: Optional[_NVIDIAClient] = None,
        list_all: bool = False,
        filter: Optional[str] = None,
        **kwargs: Any,
    ) -> List[Model]:
        """Map the available models that can be invoked. Callable from class"""
        if mode is not None:
            warn_deprecated(
                name="mode", since="0.0.17", removal="0.1.0", alternative="`base_url`"
            )
        self = client or cls(**kwargs)
        nveclient = self.client
        nveclient.reset_method_cache()
        out = sorted(
            [
                Model(id=k.replace("playground_", ""), path=v, **MODEL_SPECS.get(k, {}))
                for k, v in nveclient.available_models.items()
            ],
            key=lambda x: f"{x.client or 'Z'}{x.id}{cls}",
        )
        # nim model listing does not provide the type and we cannot know
        # the model name ahead of time to guess the type.
        # so we need to list all models.
        if mode == "nim" or not self.is_hosted:
            list_all = True
        if not filter:
            filter = cls.__name__
        if not list_all:
            out = [m for m in out if m.client == filter]
        return out

    @deprecated(since="0.0.15", removal="0.1.0")
    def get_model_details(self, model: Optional[str] = None) -> dict:
        """Get more meta-details about a model retrieved by a given name"""
        if model is None:
            model = self.model
        model_key = self.client._get_invoke_url(model).split("/")[-1]
        known_fns = self.client.available_functions
        fn_spec = [f for f in known_fns if f.get("id") == model_key][0]
        return fn_spec

    @deprecated(since="0.0.15", removal="0.1.0")
    def get_binding_model(self) -> Optional[str]:
        """Get the model to bind to the client as default payload argument"""
        # if a model is configured with a model_name, always use that
        # todo: move from search of available_models to a Model property
        matches = [model for model in self.available_models if model.id == self.model]
        if matches:
            if matches[0].model_name:
                return matches[0].model_name
        if self.curr_mode == "nvidia":
            return ""
        return self.model

    @deprecated(
        since="0.0.17",
        removal="0.1.0",
        alternative="`base_url` in constructor",
    )
    def mode(
        self,
        mode: Optional[_MODE_TYPE] = "nvidia",
        base_url: Optional[str] = None,
        model: Optional[str] = None,
        api_key: Optional[str] = None,
        infer_path: Optional[str] = None,
        models_path: Optional[str] = "{base_url}/models",
        force_mode: bool = False,
        force_clone: bool = True,
        **kwargs: Any,
    ) -> Any:  # todo: in python 3.11+ this should be typing.Self
        """Deprecated: pass `base_url=...` to constructor instead."""
        if isinstance(self, str):
            raise ValueError("Please construct the model before calling mode()")
        out = self if not force_clone else deepcopy(self)

        if mode is None:
            return out

        out.model = model or out.model

        if base_url and not force_mode:
            ## If a user tries to set base_url, assume custom openapi unless forced
            mode = "nim"

        if mode == "nvidia":
            key_var = "NVIDIA_API_KEY"
            if not api_key or not api_key.startswith("nvapi-"):
                api_key = os.getenv(key_var) or out.client.api_key.get_secret_value()
            if not api_key.startswith("nvapi-"):
                raise ValueError(f"No {key_var} in env/fed as api_key. (nvapi-...)")

        out.curr_mode = mode
        if api_key:
            out.client.api_key = SecretStr(api_key)

        nvcf_base = "https://api.nvcf.nvidia.com/v2/nvcf"  ## NVCF Main URL
        nvcf_infer = "{base_url}/pexec/functions/{model_id}"  ## Inference endpoints
        nvcf_status = "{base_url}/pexec/status/{request_id}"  ## 202 wait handle
        nvcf_models = "{base_url}/functions"  ## Model listing

        if mode == "nvidia":
            ## Classic support for nvcf-backed foundation model endpoints.
            out.client.base_url = base_url or nvcf_base
            out.client.endpoints = {
                "infer": nvcf_infer,  ## Per-model inference
                "status": nvcf_status,  ## 202 wait handle
                "models": nvcf_models,  ## Model listing
            }

        elif mode == "nim":
            ## OpenAPI-style specs to connect to NeMo Inference Microservices etc.
            ## Most generic option, requires specifying base_url
            assert base_url, "Base URL must be specified for nim mode"
            out.client.base_url = base_url
            out.client.endpoints["infer"] = infer_path or out.infer_endpoint
            out.client.endpoints["models"] = models_path or "{base_url}/models"

        else:
            options = ["nvidia", "nim"]
            raise ValueError(f"Unknown mode: `{mode}`. Expected one of {options}.")

        out.client.reset_method_cache()

        return out



"""
**LangChain NVIDIA AI Foundation Model Playground Integration**

This comprehensive module integrates NVIDIA's state-of-the-art AI Foundation Models, featuring advanced models for conversational AI and semantic embeddings, into the LangChain framework. It provides robust classes for seamless interaction with NVIDIA's AI models, particularly tailored for enriching conversational experiences and enhancing semantic understanding in various applications.

**Features:**

1. **Chat Models (`ChatNVIDIA`):** This class serves as the primary interface for interacting with NVIDIA's Foundation chat models. Users can effortlessly utilize NVIDIA's advanced models like 'Mistral' to engage in rich, context-aware conversations, applicable across diverse domains from customer support to interactive storytelling.

2. **Semantic Embeddings (`NVIDIAEmbeddings`):** The module offers capabilities to generate sophisticated embeddings using NVIDIA's AI models. These embeddings are instrumental for tasks like semantic analysis, text similarity assessments, and contextual understanding, significantly enhancing the depth of NLP applications.

**Installation:**

Install this module easily using pip:

```python
pip install langchain-nvidia-ai-endpoints
```

## Utilizing Chat Models:

After setting up the environment, interact with NVIDIA AI Foundation models:
```python
from langchain_nvidia_ai_endpoints import ChatNVIDIA

ai_chat_model = ChatNVIDIA(model="meta/llama2-70b")
response = ai_chat_model.invoke("Tell me about the LangChain integration.")
```

# Generating Semantic Embeddings:

Use NVIDIA's models for creating embeddings, useful in various NLP tasks:

```python
from langchain_nvidia_ai_endpoints import NVIDIAEmbeddings

embed_model = NVIDIAEmbeddings(model="nvolveqa_40k")
embedding_output = embed_model.embed_query("Exploring AI capabilities.")
```
"""  # noqa: E501

from langchain_nvidia_ai_endpoints.chat_models import ChatNVIDIA
from langchain_nvidia_ai_endpoints.embeddings import NVIDIAEmbeddings
from langchain_nvidia_ai_endpoints.reranking import NVIDIARerank

__all__ = ["ChatNVIDIA", "NVIDIAEmbeddings", "NVIDIARerank"]



from typing import Optional

from langchain_core.pydantic_v1 import BaseModel


class Model(BaseModel):
    id: str
    model_type: Optional[str] = None
    api_type: Optional[str] = None
    model_name: Optional[str] = None
    client: Optional[str] = None
    path: str


MODEL_SPECS = {
    "playground_smaug_72b": {"model_type": "chat", "api_type": "aifm"},
    "playground_kosmos_2": {
        "model_type": "image_in",
        "api_type": "aifm",
        "alternative": "microsoft/kosmos-2",
    },
    "playground_llama2_70b": {
        "model_type": "chat",
        "api_type": "aifm",
        "alternative": "meta/llama2-70b",
    },
    "playground_gemma_7b": {
        "model_type": "chat",
        "api_type": "aifm",
        "alternative": "google/gemma-7b",
    },
    "playground_mistral_7b": {
        "model_type": "chat",
        "api_type": "aifm",
        "alternative": "mistralai/mistral-7b-instruct-v0.2",
    },
    "playground_mamba_chat": {"model_type": "chat", "api_type": "aifm"},
    "playground_phi2": {
        "model_type": "chat",
        "api_type": "aifm",
        "alternative": "microsoft/phi-3-mini-128k-instruct",
    },
    "playground_sdxl": {"model_type": "image_out", "api_type": "aifm"},
    "playground_nv_llama2_rlhf_70b": {"model_type": "chat", "api_type": "aifm"},
    "playground_neva_22b": {
        "model_type": "image_in",
        "api_type": "aifm",
        "alternative": "nvidia/neva-22b",
    },
    "playground_yi_34b": {"model_type": "chat", "api_type": "aifm"},
    "playground_nemotron_steerlm_8b": {"model_type": "chat", "api_type": "aifm"},
    "playground_cuopt": {"model_type": "cuopt", "api_type": "aifm"},
    "playground_llama_guard": {"model_type": "classifier", "api_type": "aifm"},
    "playground_starcoder2_15b": {"model_type": "completion", "api_type": "aifm"},
    "playground_deplot": {
        "model_type": "image_in",
        "api_type": "aifm",
        "alternative": "google/deplot",
    },
    "playground_llama2_code_70b": {
        "model_type": "chat",
        "api_type": "aifm",
        "alternative": "meta/codelama-70b",
    },
    "playground_gemma_2b": {
        "model_type": "chat",
        "api_type": "aifm",
        "alternative": "google/gemma-2b",
    },
    "playground_seamless": {"model_type": "translation", "api_type": "aifm"},
    "playground_mixtral_8x7b": {
        "model_type": "chat",
        "api_type": "aifm",
        "alternative": "mistralai/mixtral-8x7b-instruct-v0.1",
    },
    "playground_fuyu_8b": {
        "model_type": "image_in",
        "api_type": "aifm",
        "alternative": "adept/fuyu-8b",
    },
    "playground_llama2_code_34b": {
        "model_type": "chat",
        "api_type": "aifm",
        "alternative": "meta/codellama-70b",
    },
    "playground_llama2_code_13b": {
        "model_type": "chat",
        "api_type": "aifm",
        "alternative": "meta/codellama-70b",
    },
    "playground_steerlm_llama_70b": {"model_type": "chat", "api_type": "aifm"},
    "playground_clip": {"model_type": "similarity", "api_type": "aifm"},
    "playground_llama2_13b": {
        "model_type": "chat",
        "api_type": "aifm",
        "alternative": "meta/llama2-70b",
    },
}

MODEL_SPECS.update(
    {
        "ai-codellama-70b": {"model_type": "chat", "model_name": "meta/codellama-70b"},
        "ai-embed-qa-4": {"model_type": "embedding", "model_name": "NV-Embed-QA"},
        "ai-fuyu-8b": {"model_type": "image_in", "model_name": "adept/fuyu-8b"},
        "ai-gemma-7b": {"model_type": "chat", "model_name": "google/gemma-7b"},
        "ai-google-deplot": {"model_type": "image_in", "model_name": "google/deplot"},
        "ai-llama2-70b": {"model_type": "chat", "model_name": "meta/llama2-70b"},
        "ai-microsoft-kosmos-2": {
            "model_type": "image_in",
            "model_name": "microsoft/kosmos-2",
        },
        "ai-mistral-7b-instruct-v2": {
            "model_type": "chat",
            "model_name": "mistralai/mistral-7b-instruct-v0.2",
        },
        "ai-mixtral-8x7b-instruct": {
            "model_type": "chat",
            "model_name": "mistralai/mixtral-8x7b-instruct-v0.1",
        },
        "ai-neva-22b": {"model_type": "image_in", "model_name": "nvidia/neva-22b"},
        "ai-rerank-qa-mistral-4b": {
            "model_type": "ranking",
            "model_name": "nv-rerank-qa-mistral-4b:1",  # nvidia/rerank-qa-mistral-4b
        },
        # 'ai-sdxl-turbo': {'model_type': 'image_out'},
        # 'ai-stable-diffusion-xl-base': {'model_type': 'iamge_out'},
        "ai-codegemma-7b": {"model_type": "chat", "model_name": "google/codegemma-7b"},
        "ai-recurrentgemma-2b": {
            "model_type": "chat",
            "model_name": "google/recurrentgemma-2b",
        },
        "ai-gemma-2b": {"model_type": "chat", "model_name": "google/gemma-2b"},
        "ai-mistral-large": {
            "model_type": "chat",
            "model_name": "mistralai/mistral-large",
        },
        "ai-mixtral-8x22b": {
            "model_type": "completion",
            "model_name": "mistralai/mixtral-8x22b-v0.1",
        },
        "ai-mixtral-8x22b-instruct": {
            "model_type": "chat",
            "model_name": "mistralai/mixtral-8x22b-instruct-v0.1",
        },
        "ai-llama3-8b": {"model_type": "chat", "model_name": "meta/llama3-8b-instruct"},
        "ai-llama3-70b": {
            "model_type": "chat",
            "model_name": "meta/llama3-70b-instruct",
        },
        "ai-phi-3-mini": {
            "model_type": "chat",
            "model_name": "microsoft/phi-3-mini-128k-instruct",
        },
        "ai-arctic": {"model_type": "chat", "model_name": "snowflake/arctic"},
        "ai-dbrx-instruct": {
            "model_type": "chat",
            "model_name": "databricks/dbrx-instruct",
        },
        "ai-phi-3-mini-4k": {
            "model_type": "chat",
            "model_name": "microsoft/phi-3-mini-4k-instruct",
        },
        "ai-seallm-7b": {"model_type": "chat", "model_name": "seallms/seallm-7b-v2.5"},
        "ai-sea-lion-7b-instruct": {
            "model_type": "chat",
            "model_name": "aisingapore/sea-lion-7b-instruct",
        },
        "ai-google-paligemma": {
            "model_type": "image_in",
            "model_name": "google/paligemma",
        },
        "ai-arctic-embed-l": {
            "model_type": "embedding",
            "model_name": "snowflake/arctic-embed-l",
        },
        "ai-phi-3-small-8k-instruct": {
            "model_type": "chat",
            "model_name": "microsoft/phi-3-small-8k-instruct",
        },
        "ai-phi-3-small-128k-instruct": {
            "model_type": "chat",
            "model_name": "microsoft/phi-3-small-128k-instruct",
        },
        "ai-phi-3-medium-4k-instruct": {
            "model_type": "chat",
            "model_name": "microsoft/phi-3-medium-4k-instruct",
        },
        "ai-phi-3-vision-128k-instruct": {
            "model_type": "image_in",
            "model_name": "microsoft/phi-3-vision-128k-instruct",
        },
    }
)

client_map = {
    "asr": "None",
    "chat": "ChatNVIDIA",
    "classifier": "None",
    "completion": "NVIDIA",
    "cuopt": "None",
    "embedding": "NVIDIAEmbeddings",
    "image_in": "ChatNVIDIA",
    "image_out": "ImageGenNVIDIA",
    "qa": "ChatNVIDIA",
    "similarity": "None",
    "translation": "None",
    "tts": "None",
    "ranking": "NVIDIARerank",
}

MODEL_SPECS = {
    k: {**v, "client": client_map[v["model_type"]]} for k, v in MODEL_SPECS.items()
}
# MODEL_SPEC database should have both function and model names
MODEL_SPECS.update(
    {v["model_name"]: v for v in MODEL_SPECS.values() if "model_name" in v}
)



from __future__ import annotations

from typing import (
    Any,
    AsyncIterator,
    Dict,
    Iterator,
    List,
    Optional,
    Sequence,
    Union,
)

from langchain_core.callbacks.manager import (
    AsyncCallbackManagerForLLMRun,
    CallbackManagerForLLMRun,
)
from langchain_core.language_models import LLM
from langchain_core.outputs import GenerationChunk
from langchain_core.pydantic_v1 import Field

from langchain_nvidia_ai_endpoints import _common as nvidia_ai_endpoints

_CallbackManager = Union[AsyncCallbackManagerForLLMRun, CallbackManagerForLLMRun]


"""
### Code Generation

These models accept the same arguments and input structure as regular chat models, but
they tend to perform better on code-genreation and structured code tasks. An example
of this is `llama2_code_70b`.

```
prompt = ChatPromptTemplate.from_messages(
    [
        (
            "system",
            "You are an expert coding AI. Respond only in valid python;
            no narration whatsoever.",
        ),
        ("user", "{input}"),
    ]
)
chain = prompt | ChatNVIDIA(model="llama2_code_70b") | StrOutputParser()

for txt in chain.stream({"input": "How do I solve this fizz buzz problem?"}):
    print(txt, end="")
```

In addition, the [**StarCoder2**]
(https://catalog.ngc.nvidia.com/orgs/nvidia/teams/ai-foundation/models/starcoder2-15b)
model also supports code generation, but subscribes to a regular completion API. For
this, you should use the LLM-style `NVIDIA` class:

```
from langchain_nvidia_ai_endpoints import NVIDIA

starcoder = NVIDIA(model="starcoder2_15b", stop=["```"])

# print(chain.invoke("Here is my implementation of fizzbuzz:\n```python\n", stop="```"))
for txt in starcoder.stream("Here is my implementation of fizzbuzz:\n```python\n"):
    print(txt, end="")
```
"""


class NVIDIA(nvidia_ai_endpoints._NVIDIAClient, LLM):
    """NVIDIA chat model.

    Example:
        .. code-block:: python

            from langchain_nvidia_ai_endpoints import ChatNVIDIA


            model = NVIDIA(model="starcoder2_15b")
            response = model.invoke("Here is my fizzbuzz code:\n```python\n")
    """

    _default_model: str = "starcoder2_15b"
    infer_endpoint: str = Field("{base_url}/completions")
    model: str = Field(_default_model, description="Name of the model to invoke")
    temperature: Optional[float] = Field(description="Sampling temperature in [0, 1]")
    max_tokens: Optional[int] = Field(description="Maximum # of tokens to generate")
    top_p: Optional[float] = Field(description="Top-p for distribution sampling")
    frequency_penalty: Optional[float] = Field(description="Frequency penalty")
    presence_penalty: Optional[float] = Field(description="Presence penalty")
    seed: Optional[int] = Field(description="The seed for deterministic results")
    bad: Optional[Sequence[str]] = Field(description="Bad words to avoid (cased)")
    stop: Optional[Sequence[str]] = Field(description="Stop words (cased)")
    labels: Optional[Dict[str, float]] = Field(description="Steering parameters")
    streaming: bool = Field(True)

    @property
    def _llm_type(self) -> str:
        """Return type of NVIDIA AI Foundation Model Interface."""
        return "nvidia-ai-playground"

    def _call(
        self,
        prompt: str,
        stop: Optional[List[str]] = None,
        run_manager: Optional[CallbackManagerForLLMRun] = None,
        **kwargs: Any,
    ) -> str:
        """Invoke on a single list of chat messages."""
        response = self.get_generation(prompt=prompt, stop=stop, **kwargs)
        output = self.custom_postprocess(response)
        return output

    def _stream(
        self,
        prompt: str,
        stop: Optional[List[str]] = None,
        run_manager: Optional[CallbackManagerForLLMRun] = None,
        **kwargs: Any,
    ) -> Iterator[GenerationChunk]:
        """Allows streaming to model!"""
        for response in self.get_stream(prompt=prompt, stop=stop, **kwargs):
            self._set_callback_out(response, run_manager)
            chunk = GenerationChunk(text=self.custom_postprocess(response))
            yield chunk
            if run_manager:
                run_manager.on_llm_new_token(chunk.text, chunk=chunk)

    async def _astream(
        self,
        prompt: str,
        stop: Optional[List[str]] = None,
        run_manager: Optional[AsyncCallbackManagerForLLMRun] = None,
        **kwargs: Any,
    ) -> AsyncIterator[GenerationChunk]:
        async for response in self.get_astream(prompt=prompt, stop=stop, **kwargs):
            self._set_callback_out(response, run_manager)
            chunk = GenerationChunk(text=self.custom_postprocess(response))
            yield chunk
            if run_manager:
                await run_manager.on_llm_new_token(chunk.text, chunk=chunk)

    def _set_callback_out(
        self,
        result: dict,
        run_manager: Optional[_CallbackManager],
    ) -> None:
        result.update({"model_name": self.model})
        if run_manager:
            for cb in run_manager.handlers:
                if hasattr(cb, "llm_output"):
                    cb.llm_output = result

    def custom_postprocess(self, msg: dict) -> str:
        if "content" in msg:
            return msg["content"]
        elif "b64_json" in msg:
            return msg["b64_json"]
        return str(msg)

    ######################################################################################
    ## Core client-side interfaces

    def get_generation(
        self,
        prompt: str,
        **kwargs: Any,
    ) -> dict:
        """Call to client generate method with call scope"""
        stop = kwargs["stop"] = kwargs.get("stop") or self.stop
        payload = self.get_payload(prompt=prompt, stream=False, **kwargs)
        out = self.client.get_req_generation(self.model, stop=stop, payload=payload)
        return out

    def get_stream(
        self,
        prompt: str,
        **kwargs: Any,
    ) -> Iterator:
        """Call to client stream method with call scope"""
        stop = kwargs["stop"] = kwargs.get("stop") or self.stop
        payload = self.get_payload(prompt=prompt, stream=True, **kwargs)
        return self.client.get_req_stream(self.model, stop=stop, payload=payload)

    def get_astream(
        self,
        prompt: str,
        **kwargs: Any,
    ) -> AsyncIterator:
        """Call to client astream methods with call scope"""
        stop = kwargs["stop"] = kwargs.get("stop") or self.stop
        payload = self.get_payload(prompt=prompt, stream=True, **kwargs)
        return self.client.get_req_astream(self.model, stop=stop, payload=payload)

    def get_payload(self, prompt: str, **kwargs: Any) -> dict:
        """Generates payload for the _NVIDIAClient API to send to service."""
        attr_kwargs = {
            "temperature": self.temperature,
            "max_tokens": self.max_tokens,
            "top_p": self.top_p,
            "seed": self.seed,
            "bad": self.bad,
            "stop": self.stop,
            "labels": self.labels,
        }
        if self.get_binding_model():
            attr_kwargs["model"] = self.get_binding_model()
        attr_kwargs = {k: v for k, v in attr_kwargs.items() if v is not None}
        new_kwargs = {**attr_kwargs, **kwargs}
        return self.prep_payload(prompt=prompt, **new_kwargs)

    def prep_payload(self, prompt: str, **kwargs: Any) -> dict:
        """Prepares a message or list of messages for the payload"""
        if kwargs.get("stop") is None:
            kwargs.pop("stop")
        return {"prompt": prompt, **kwargs}



from __future__ import annotations

from typing import Any, Generator, List, Optional, Sequence

from langchain_core._api import deprecated, warn_deprecated
from langchain_core.callbacks.manager import Callbacks
from langchain_core.documents import Document
from langchain_core.documents.compressor import BaseDocumentCompressor
from langchain_core.pydantic_v1 import BaseModel, Field, PrivateAttr

from ._common import _MODE_TYPE, _NVIDIAClient
from ._statics import Model


class Ranking(BaseModel):
    index: int
    logit: float


class NVIDIARerank(BaseDocumentCompressor):
    """
    LangChain Document Compressor that uses the NVIDIA NeMo Retriever Reranking API.
    """

    class Config:
        validate_assignment = True

    _client: _NVIDIAClient = PrivateAttr(_NVIDIAClient)

    _default_batch_size: int = 32
    _deprecated_model: str = "ai-rerank-qa-mistral-4b"
    _default_model_name: str = "nv-rerank-qa-mistral-4b:1"

    top_n: int = Field(5, ge=0, description="The number of documents to return.")
    model: str = Field(
        _default_model_name, description="The model to use for reranking."
    )
    max_batch_size: int = Field(
        _default_batch_size, ge=1, description="The maximum batch size."
    )
    _is_hosted: bool = PrivateAttr(True)

    def __init__(self, **kwargs: Any):
        """
        Create a new NVIDIARerank document compressor.

        This class provides access to a NVIDIA NIM for reranking. By default, it
        connects to a hosted NIM, but can be configured to connect to a local NIM
        using the `base_url` parameter. An API key is required to connect to the
        hosted NIM.

        Args:
            model (str): The model to use for reranking.
            nvidia_api_key (str): The API key to use for connecting to the hosted NIM.
            api_key (str): Alternative to nvidia_api_key.
            base_url (str): The base URL of the NIM to connect to.

        API Key:
        - The recommended way to provide the API key is through the `NVIDIA_API_KEY`
            environment variable.
        """
        super().__init__(**kwargs)
        self._client = _NVIDIAClient(
            model=self.model,
            api_key=kwargs.get("nvidia_api_key", kwargs.get("api_key", None)),
        )
        if base_url := kwargs.get("base_url", None):
            # todo: detect if the base_url points to hosted NIM, this depends on
            #       moving from NVCF inference to API Catalog inference
            self._is_hosted = False
            self._client.client.base_url = base_url
            self._client.client.endpoints["infer"] = "{base_url}/ranking"
            self._client.client.endpoints = {
                "infer": "{base_url}/ranking",
                "status": None,
                "models": None,
            }

    @property
    def available_models(self) -> List[Model]:
        """
        Get a list of available models that work with NVIDIARerank.
        """
        if self._client.curr_mode == "nim" or not self._is_hosted:
            # local NIM supports a single model and no /models endpoint
            models = [
                Model(
                    id=NVIDIARerank._default_model_name,
                    model_name=NVIDIARerank._default_model_name,
                    model_type="ranking",
                    client="NVIDIARerank",
                    path="magic",
                ),
                Model(
                    id=NVIDIARerank._deprecated_model,
                    model_name=NVIDIARerank._default_model_name,
                    model_type="ranking",
                    client="NVIDIARerank",
                    path="magic",
                ),
            ]
        else:
            models = self._client.get_available_models(
                client=self._client,
                filter=self.__class__.__name__,
            )
        return models

    @classmethod
    def get_available_models(
        cls,
        mode: Optional[_MODE_TYPE] = None,
        list_all: bool = False,
        **kwargs: Any,
    ) -> List[Model]:
        """
        Get a list of available models. These models will work with the NVIDIARerank
        interface.

        Use the mode parameter to specify the mode to use. See the docs for mode()
        to understand additional keyword arguments required when setting mode.

        It is possible to get a list of all models, including those that are not
        chat models, by setting the list_all parameter to True.
        """
        if mode is not None:
            warn_deprecated(since="0.0.17", removal="0.1.0", alternative="`base_url`")
        self = cls(**kwargs).mode(mode=mode, **kwargs)
        if mode == "nim" or not self._is_hosted:
            # ignoring list_all because there is one
            models = self.available_models
        else:
            models = self._client.get_available_models(
                mode=mode,
                list_all=list_all,
                client=self._client,
                filter=cls.__name__,
                **kwargs,
            )
        return models

    @deprecated(
        since="0.0.17",
        removal="0.1.0",
        alternative="`base_url` to constructor",
    )
    def mode(
        self,
        mode: Optional[_MODE_TYPE] = "nvidia",
        base_url: Optional[str] = None,
        model: Optional[str] = None,
        api_key: Optional[str] = None,
        **kwargs: Any,
    ) -> NVIDIARerank:
        """
        Deprecated: use NVIDIARerank(base_url=...) instead.
        """
        # set a default base_url for nim mode
        if not base_url and mode == "nim":
            base_url = "http://localhost:1976/v1"
        self._client = self._client.mode(
            mode=mode,
            base_url=base_url,
            model=model,
            api_key=api_key,
            infer_path="{base_url}/ranking",
            **kwargs,
        )
        self.model = self._client.model
        return self

    # todo: batching when len(documents) > endpoint's max batch size
    def _rank(self, documents: List[str], query: str) -> List[Ranking]:
        response = self._client.client.get_req(
            model_name=self.model,
            payload={
                "model": "nv-rerank-qa-mistral-4b:1",
                "query": {"text": query},
                "passages": [{"text": passage} for passage in documents],
            },
            endpoint="infer",
        )
        if response.status_code != 200:
            response.raise_for_status()
        # todo: handle errors
        rankings = response.json()["rankings"]
        # todo: callback support
        return [Ranking(**ranking) for ranking in rankings[: self.top_n]]

    def compress_documents(
        self,
        documents: Sequence[Document],
        query: str,
        callbacks: Optional[Callbacks] = None,
    ) -> Sequence[Document]:
        """
        Compress documents using the NVIDIA NeMo Retriever Reranking microservice API.

        Args:
            documents: A sequence of documents to compress.
            query: The query to use for compressing the documents.
            callbacks: Callbacks to run during the compression process.

        Returns:
            A sequence of compressed documents.
        """
        if len(documents) == 0 or self.top_n < 1:
            return []

        def batch(ls: list, size: int) -> Generator[List[Document], None, None]:
            for i in range(0, len(ls), size):
                yield ls[i : i + size]

        doc_list = list(documents)
        results = []
        for doc_batch in batch(doc_list, self.max_batch_size):
            rankings = self._rank(
                query=query, documents=[d.page_content for d in doc_batch]
            )
            for ranking in rankings:
                doc = doc_batch[ranking.index]
                doc.metadata["relevance_score"] = ranking.logit
                results.append(doc)

        # if we batched, we need to sort the results
        if len(doc_list) > self.max_batch_size:
            results.sort(key=lambda x: x.metadata["relevance_score"], reverse=True)

        return results[: self.top_n]



"""Embeddings Components Derived from NVEModel/Embeddings"""

import warnings
from typing import List, Literal, Optional

from langchain_core.embeddings import Embeddings
from langchain_core.outputs.llm_result import LLMResult
from langchain_core.pydantic_v1 import Field, validator

from langchain_nvidia_ai_endpoints._common import _NVIDIAClient
from langchain_nvidia_ai_endpoints.callbacks import usage_callback_var

from ._statics import MODEL_SPECS


class NVIDIAEmbeddings(_NVIDIAClient, Embeddings):
    """
    Client to NVIDIA embeddings models.

    Fields:
    - model: str, the name of the model to use
    - truncate: "NONE", "START", "END", truncate input text if it exceeds the model's
        maximum token length. Default is "NONE", which raises an error if an input is
        too long.
    """

    _default_model: str = "NV-Embed-QA"
    _default_max_batch_size: int = 50
    infer_endpoint: str = Field("{base_url}/embeddings")
    model: str = Field(_default_model, description="Name of the model to invoke")
    truncate: Literal["NONE", "START", "END"] = Field(
        default="NONE",
        description=(
            "Truncate input text if it exceeds the model's maximum token length. "
            "Default is 'NONE', which raises an error if an input is too long."
        ),
    )
    max_length: int = Field(2048, ge=1, le=2048)
    max_batch_size: int = Field(default=_default_max_batch_size)
    model_type: Optional[Literal["passage", "query"]] = Field(
        None, description="The type of text to be embedded."
    )

    # indicate to user that max_length is deprecated when passed as an argument to
    # NVIDIAEmbeddings' constructor, e.g. NVIDIAEmbeddings(max_length=...). this
    # does not warning on assignment, e.g. embedder.max_length = ...
    # todo: fix _NVIDIAClient.validate_client and enable Config.validate_assignment
    @validator("max_length")
    def deprecated_max_length(cls, value: int) -> int:
        """Deprecate the max_length field."""
        warnings.warn(
            "The max_length field is deprecated. Use the 'truncate' instead.",
            DeprecationWarning,
        )
        return value

    # todo: fix _NVIDIAClient.validate_client and enable Config.validate_assignment
    @validator("model")
    def aifm_deprecated(cls, value: str) -> str:
        """All AI Foundataion Models are deprecate, use API Catalog models instead."""
        for model in [value, f"playground_{value}"]:
            if model in MODEL_SPECS and MODEL_SPECS[model].get("api_type") == "aifm":
                alternative = MODEL_SPECS[model].get(
                    "alternative", NVIDIAEmbeddings._default_model
                )
                warnings.warn(
                    f"{value} is deprecated. Try {alternative} instead.",
                    DeprecationWarning,
                )
        return value

    def _embed(
        self, texts: List[str], model_type: Literal["passage", "query"]
    ) -> List[List[float]]:
        """Embed a single text entry to either passage or query type"""
        # AI Foundation Model API -
        #  input: str | list[str]              -- <= 2048 characters, <= 50 inputs
        #  model: "query" | "passage"          -- type of input text to be embedded
        #  encoding_format: "float" | "base64"
        # API Catalog API -
        #  input: str | list[str]              -- char limit depends on model
        #  model: str                          -- model name, e.g. NV-Embed-QA
        #  encoding_format: "float" | "base64"
        #  input_type: "query" | "passage"
        #  user: str                           -- ignored
        #  truncate: "NONE" | "START" | "END"  -- default "NONE", error raised if
        #                                         an input is too long
        # todo: remove the playground aliases
        model_name = self.model
        if model_name not in MODEL_SPECS:
            if f"playground_{model_name}" in MODEL_SPECS:
                model_name = f"playground_{model_name}"
        if MODEL_SPECS.get(model_name, {}).get("api_type", None) == "aifm":
            payload = {
                "input": texts,
                "model": model_type,
                "encoding_format": "float",
            }
        else:  # default to the API Catalog API
            payload = {
                "input": texts,
                "model": self.get_binding_model() or self.model,
                "encoding_format": "float",
                "input_type": model_type,
            }
            if self.truncate:
                payload["truncate"] = self.truncate

        response = self.client.get_req(
            model_name=self.model,
            payload=payload,
            endpoint="infer",
        )
        response.raise_for_status()
        result = response.json()
        data = result.get("data", result)
        if not isinstance(data, list):
            raise ValueError(f"Expected data with a list of embeddings. Got: {data}")
        embedding_list = [(res["embedding"], res["index"]) for res in data]
        self._invoke_callback_vars(result)
        return [x[0] for x in sorted(embedding_list, key=lambda x: x[1])]

    def embed_query(self, text: str) -> List[float]:
        """Input pathway for query embeddings."""
        return self._embed([text], model_type=self.model_type or "query")[0]

    def embed_documents(self, texts: List[str]) -> List[List[float]]:
        """Input pathway for document embeddings."""
        if not isinstance(texts, list) or not all(
            isinstance(text, str) for text in texts
        ):
            raise ValueError(f"`texts` must be a list of strings, given: {repr(texts)}")

        # From https://catalog.ngc.nvidia.com/orgs/nvidia/teams/ai-foundation/models/nvolve-40k/documentation
        # The input must not exceed the 2048 max input characters and inputs above 512
        # model tokens will be truncated. The input array must not exceed 50 input
        #  strings.
        all_embeddings = []
        for i in range(0, len(texts), self.max_batch_size):
            batch = texts[i : i + self.max_batch_size]
            truncated = [
                text[: self.max_length] if len(text) > self.max_length else text
                for text in batch
            ]
            all_embeddings.extend(
                self._embed(truncated, model_type=self.model_type or "passage")
            )
        return all_embeddings

    def _invoke_callback_vars(self, response: dict) -> None:
        """Invoke the callback context variables if there are any."""
        callback_vars = [
            usage_callback_var.get(),
        ]
        llm_output = {**response, "model_name": self.model}
        result = LLMResult(generations=[[]], llm_output=llm_output)
        for cb_var in callback_vars:
            if cb_var:
                cb_var.on_llm_end(result)



"""Embeddings Components Derived from NVEModel/Embeddings"""
import base64
from io import BytesIO
from typing import Any, List, Optional

import requests
from langchain_core.callbacks.manager import CallbackManagerForLLMRun
from langchain_core.language_models import LLM
from langchain_core.pydantic_v1 import Field
from langchain_core.runnables import Runnable, RunnableLambda
from PIL import Image

from langchain_nvidia_ai_endpoints._common import _NVIDIAClient

"""
## Image Generation Models

Due to the similarity of the underlying API, a selection of **Image Generation Models**
can be supported using the LLM interface. One example is the [**Stable Diffusion XL**]
(https://catalog.ngc.nvidia.com/orgs/nvidia/teams/ai-foundation/models/sdxl/api) model,
which expects a prompt and some other arguments as input and produces an image (passed
back as a b64-encoded string).

```
from langchain_nvidia_ai_endpoints import ImageGenNVIDIA

ImageGenNVIDIA.get_available_models()  ## Only shows models supported by ImageGenNVIDIA

sdxl = ImageGenNVIDIA(model="sdxl")

out = sdxl.invoke("A picture of a cat, aesthetic")
out[:100] + "..."
```

To see the image, we can either convert it manually or use the built-in conversion
utility

```
from langchain_nvidia_ai_endpoints.image_gen import ImageParser

# from io import BytesIO
# from PIL import Image
# import base64

# Image.open(BytesIO(base64.decodebytes(bytes(out[10:], "utf-8"))))
ImageParser().invoke(out)  ## Runnable that does it all for you
```

In addition to the prompt, we can do a bit of hyperparameter tweaking and add some
negative prompt components that we'd like to avoid. We will also use the `ImageParser`
runnable automatically by calling the `.as_pil()` method.

```
sdxl = ImageGenNVIDIA(
    inference_steps = 100,
    negative_prompt = "ugly,bad eyes,low-res,crooked nose, smudged, painted",
)

sdxl.as_pil().invoke("A picture of a big green dog, futuristic cyberpunk")
```

Note that under the hood, `as_pil` returns a merger of the `ImageGenModel` object with
the `ImageParser` output parser. As a result, you may have trouble interacting with the
aggregation. Note that you can reference the first half of the pipeline via `.first` or
similar.

## Example of image generation with OpenAI and DALL-E

```
from langchain_nvidia_ai_endpoints import ImageGenNVIDIA
from getpass import getpass
import os

if not os.environ.get("OPENAI_API_KEY", "").startswith("sk-"):
    os.environ["OPENAI_API_KEY"] = getpass("Enter your OPENAI Key: ")

llm = ImageGenNVIDIA().mode("openai")
llm.available_models
```

```
from langchain_nvidia_ai_endpoints import ImageGenNVIDIA

dalle = ImageGenNVIDIA().mode("openai", model="dall-e-3")

def payload_fn(d):
    if d:
        drop_keys = ["guidance_scale", "seed", "negative_prompt", "sampler"]
        d = {k: v for k, v in d.items() if k not in drop_keys}
    return d

dalle.client.payload_fn = payload_fn

dalle.as_pil().invoke("City skyline, neon green tint, aesthetic futuristic realistic")
print(dalle.client.last_inputs['json'])
dalle.client.last_response.json()
```
"""


def _get_pil_from_response(data: str) -> Image.Image:
    if data.startswith("url: "):
        body = requests.get(data[4:], stream=True).raw
    elif data.startswith("b64_json: "):
        body = BytesIO(base64.decodebytes(bytes(data[10:], "utf-8")))
    else:
        raise ValueError(f"Invalid response format: {str(data)[:100]}")
    return Image.open(body)


def ImageParser() -> RunnableLambda[str, Image.Image]:
    return RunnableLambda(_get_pil_from_response)


class ImageGenNVIDIA(_NVIDIAClient, LLM):
    """NVIDIA's AI Foundation Retriever Question-Answering Asymmetric Model."""

    _default_model: str = "sdxl"
    infer_endpoint: str = Field("{base_url}/images/generations")
    model: str = Field(_default_model, description="Name of the model to invoke")
    negative_prompt: Optional[str] = Field(description="Sampling temperature in [0, 1]")
    sampler: Optional[str] = Field(description="Sampling strategy for process")
    guidance_scale: Optional[float] = Field(description="The scale of guidance")
    seed: Optional[int] = Field(description="The seed for deterministic results")

    @property
    def _llm_type(self) -> str:
        """Return type of NVIDIA AI Foundation Model Interface."""
        return "nvidia-image-gen-model"

    def _call(
        self,
        prompt: str,
        stop: Optional[List[str]] = None,
        run_manager: Optional[CallbackManagerForLLMRun] = None,
        **kwargs: Any,
    ) -> str:
        """Run the Image Gen Model on the given prompt and input."""
        payload = {
            "prompt": prompt,
            "negative_prompt": kwargs.get("negative_prompt", self.negative_prompt),
            "sampler": kwargs.get("sampler", self.sampler),
            "guidance_scale": kwargs.get("guidance_scale", self.guidance_scale),
            "seed": kwargs.get("seed", self.seed),
        }
        if self.get_binding_model():
            payload["model"] = self.get_binding_model()
        response = self.client.get_req(
            model_name=self.model, payload=payload, endpoint="infer"
        )
        response.raise_for_status()
        out_dict = response.json()
        if "data" in out_dict:
            out_dict = out_dict.get("data")[0]
        if "url" in out_dict:
            output = "url: {}".format(out_dict.get("url"))
        elif "b64_json" in out_dict:
            output = "b64_json: {}".format(out_dict.get("b64_json"))
        else:
            output = str(out_dict)
        return output

    def as_pil(self, **kwargs: Any) -> Runnable:
        """Returns a model that outputs a PIL image by default"""
        return self | ImageParser(**kwargs)
